// Unterprogramm fr die Speicherung des Grafikbildschirm als
// Windows-Bitmap-Datei (*.BMP) als Baustein fr eigene Programme
// Version 1.0 fr unkomprimierte Dateien
// Es werden smtliche Farbtiefen untersttzt (S/W, 16, 256, 24-Bit)
// (c) 1996 by Andreas Meile, HTL Brugg-Windisch

#include <stdio.h>
#include <grx.h>

int speichere_als_bmp(const char *d_name, // Dateiname (mit ".BMP"!)
                      int b_roehre_breite = 12, // Bildrhrenbeite in Zoll
                      int b_roehre_hoehe = 9)   // Bildrhrenhhe in Zoll
                      // => diese Vorgabewerte entsprechen einem Monitor
                      // mit 15" Bilddiagonale. Sinn dieser Werte: Korrekter
                      // Eintrag bei der Pixeldichte im BITMAPINFOHEADER
// ͻ
//  Wichtige Anmerkung bezglich Farbpalette: Da die LIBGRX   
//  mit einem Farbreservationssystem arbeitet, ist die GRX-   
//  Funktion "GrQueryColor()" nicht in der Lage, die Vorgabe- 
//  farben des Systems, wie man sie unmittelbar durch den     
//  "GrSetMode()"-Aufruf erhlt, korrekt auszulesen (es wird  
//  nur Schwarz, d.h. R=0, G=0, B=0 bei uninitialisierten     
//  Farben zurckgeliefert) => SMTLICHE verwendete Farben    
//  mssen vorher explizit mit "GrSetColor()" gesetzt werden! 
// ͼ

{
  // Prfen, ob berhaupt Grafikmodus
  if(GrCurrentMode() < GR_320_200_graphics)
    return 3; // 3=es ist berhaupt keine Grafik aktiv!

  // Analyse, welcher Bildtyp
  const int x_bild = GrScreenX();  // Bildschimparameter auslesen
  const int y_bild = GrScreenY();
  const int a_far = GrNumColors();

  short bit_pr_pix, tiefe_15_flag = (a_far <= 32768); 
  
  // Bestimmung, welcher Bildtyp

  if(a_far <= 2)
    bit_pr_pix = 1; // Schwarz/Weiss-Grafik
  else if(a_far <= 16)
    bit_pr_pix = 4; // Als 16-Farb-Palettengrafik speichern
  else if(a_far <= 256)
    bit_pr_pix = 8; // Als 256-Farb-Palettengrafik
  else
    bit_pr_pix = 24; // Als Echtfarb (bei 32768 Farben ebenfalls)

  const int a_far_bild = 1 << bit_pr_pix; // Anzahl Farben beim Bild effektiv
  const long zei_laenge = (x_bild * bit_pr_pix + 31 & -32) >> 3; // effektive
  // Zeilenlnge auf der Disk

  // Verschiedene Konstanten und Werte fr den Dateikopf
  const char kennung[2] = {'B', 'M'}; // Kennung 
  const short reserviert = 0; // reservierte Bereiche
  const long BITMAPFILEHEADER_laenge = 0x0e; // Grsse des BITMAPFILEHEADERs
  const long BITMAPINFOHEADER_laenge = 0x28; // Grsse des BITMAPINFOHEADERs
  const long bitm_dat_abst = (BITMAPFILEHEADER_laenge +
  BITMAPINFOHEADER_laenge) + (bit_pr_pix < 15 ? (a_far_bild << 2) : 0);
  // ^ SEEK-Offset zur Bitmap
  const long bild_laenge = zei_laenge * y_bild; // Bitmapdaten-Gesamtgrsse
  const long ges_laenge = bitm_dat_abst + bild_laenge; // Gesamtgrsse der
  // Datei
  const short anz_bit_pl = 1;
  const int NO_COMPRESSION = 0; // Keine Komprimierung 
  // Punktdichten (die Addition des halben Nenners dient nur dazu, um
  // eine korrekte Rundung zu bewirken
  const int x_dpm = (x_bild * 5000 + ((127 * b_roehre_breite) >> 2)) /
  (127 * b_roehre_breite);
  const int y_dpm = (y_bild * 5000 + ((127 * b_roehre_hoehe) >> 2)) /
  (127 * b_roehre_hoehe);
  const int anz_ben_wi_far = 0; // Anzahl benutzte und wichtige Farben

  // Beginn des eigentlichen Dateispeicherungsprozesses
  FILE *d_handle;

  d_handle = fopen(d_name, "wb");
  if(!d_handle)
    return 2; // Probleme beim Schreiben (ffnen) der Datei

  // Kopf speichern: BITMAPFILEHEADER
  fwrite(kennung, 1, sizeof(kennung), d_handle); // Dateikennung "BM"
  fwrite(&ges_laenge, 1, sizeof(ges_laenge), d_handle); // Gesamtdateilnge
  fwrite(&reserviert, 1, sizeof(reserviert), d_handle); // reservierte Worte
  fwrite(&reserviert, 1, sizeof(reserviert), d_handle); // reservierte Worte
  fwrite(&bitm_dat_abst, 1, sizeof(bitm_dat_abst), d_handle); // Entf. Bitmap
  // Speicherung des BITMAPINFOHEADERs
  fwrite(&BITMAPINFOHEADER_laenge, 1, sizeof(BITMAPINFOHEADER_laenge),
  d_handle); // Grsse des BITMAPINFOHEADERs
  fwrite(&x_bild, 1, sizeof(x_bild), d_handle); // Bildbreite
  fwrite(&y_bild, 1, sizeof(y_bild), d_handle); // Bildbreite
  fwrite(&anz_bit_pl, 1, sizeof(anz_bit_pl), d_handle); // Anz. Bitebenen
  fwrite(&bit_pr_pix, 1, sizeof(bit_pr_pix), d_handle); // Anz. Bit pro Pix.
  fwrite(&NO_COMPRESSION, 1, sizeof(NO_COMPRESSION), d_handle); // keine
  // Kompression
  fwrite(&bild_laenge, 1, sizeof(bild_laenge), d_handle); // Bitmapgrsse
  fwrite(&x_dpm, 1, sizeof(x_dpm), d_handle); // Horizontale Punktdichte
  fwrite(&y_dpm, 1, sizeof(y_dpm), d_handle); // Vertikale Punktdichte
  fwrite(&anz_ben_wi_far, 1, sizeof(anz_ben_wi_far), d_handle); // Anzahl
  // benutzte Farben auf 0 setzen (allgemein gebruchlich, sogar bei
  // Microsoft(R) Paintbrush der Fall!
  fwrite(&anz_ben_wi_far, 1, sizeof(anz_ben_wi_far), d_handle); // Anzahl
  // wichtige Farben => 0 => alle

  // Farbpalette speichern
  if(bit_pr_pix < 15) {
    int rot, gruen, blau;
    unsigned char farbe[4];
    farbe[3] = 0; // immer Null, da reservert
    for(int i=0; i<a_far_bild; i++) {
      GrQueryColor(i, &rot, &gruen, &blau);
      farbe[0] = blau;  // umgekehrte Reihenfolge einhalten
      farbe[1] = gruen;
      farbe[2] = rot;
      fwrite(farbe, 1, sizeof(farbe), d_handle);
    }
  }

  // Eigentliche Bitmapdaten speichern (rckwrts)
  int f_wert;
  unsigned char rgb_trip[3], dat_byte;
  for(int y = y_bild - 1; y > -1; y--) {
    for(int x = 0; x < x_bild; x++) {
      f_wert = GrPixel(x, y);
      switch(bit_pr_pix) {
        case 1 : {  // Schwarzweiss
          dat_byte = dat_byte << 1 | ((unsigned char)f_wert);
          if(!(~x & 0x7))
            fwrite(&dat_byte, 1, sizeof(dat_byte), d_handle);
          break;
        }
        case 4 : {  // 16 Farben
          dat_byte = dat_byte << 4 | ((unsigned char)f_wert);
          if(x & 0x1) {
            fwrite(&dat_byte, 1, sizeof(dat_byte), d_handle);
          }
          break;
        }
        case 8 : {  // 256 Farben
          dat_byte = f_wert;
          fwrite(&dat_byte, 1, sizeof(dat_byte), d_handle);
          break;
        }
        default : { // 24 Bit-Echtfarbe
          if(tiefe_15_flag) {  // 32K-Farbmodus
            for(register int i=0; i<3; i++) {
              rgb_trip[i] = ((unsigned char)(f_wert & 0x1f)) << 3 |
              ((unsigned char)(f_wert & 0x1c)) >> 2;
              f_wert = f_wert >> 5;
            }
          } else {  // echte 16 Millionen Farben
            for(register int i=0; i<3; i++) {
              rgb_trip[i] = f_wert;
              f_wert = f_wert >> 8;
            }
          }
          fwrite(rgb_trip, 1, sizeof(rgb_trip), d_handle);
        }
      }
    }
    // Auffllen auf die erforderliche Byteanzahl
    switch(bit_pr_pix) {
      case 1 : { // Schwarzweiss
        if(x_bild & 0x1f) {
          dat_byte = dat_byte << (-x_bild & 0x7);
          for(register int i=(39 - (x_bild & 0x1f)) >> 3; i; i--)
            fwrite(&dat_byte, 1, sizeof(dat_byte), d_handle);
        }
        break;
      }
      case 4 : { // 16 Farben
        if(x_bild & 0x7) {
          if(x_bild & 0x1)
            dat_byte = dat_byte << 4;
          for(register int i=(9 - (x_bild & 0x7)) >> 1; i; i--)
            fwrite(&dat_byte, 1, sizeof(dat_byte), d_handle);
        }
        break;
      }
      case 8 : { // 256 Farben
        if(x_bild & 0x3)
          for(register int i=(x_bild & 0x3); i<4; i++)
            fwrite(&dat_byte, 1, sizeof(dat_byte), d_handle);
        break;
      }
      default : { // Echtfarb
        if(x_bild & 0x3)
          fwrite(&rgb_trip, 1, sizeof(rgb_trip[0]) * (x_bild & 0x3),
          d_handle);
      }
    }
  }
  // Test, ob kein Schreibfehler aufgetreten ist: Die SEEK-Position muss
  // mit der berechneten Lnge bereinstimmen!
  long l_test = ftell(d_handle);
  fclose(d_handle);
  return (l_test != ges_laenge); // Schlusstest: Lnge muss identisch sein
}
